//lrt.cpp
#include "lrtMP.h"
#include "matrix.h"
#include "vector.h"
#include <iostream>
#include <cstdlib>
#include <cmath>

using std::cout; using std:: endl;

LRT::LRT(int Order, int N, double* PDat){
  order = Order;
  n = N;
  pDat = new double[n];
  memoryCheck(pDat);
  for(int i = 0; i < n ;i++)
    pDat[i] = PDat[i];
  //pointer to store Q(p + 1, tau) and Q(tau + 1, n)
  pQ = new double*[n - 2*order - 1];
  memoryCheck(pQ);
  for(int i = 0; i < (n - 2*order - 1); i++){
    pQ[i] = new double[2];
    memoryCheck(pQ[i]);
  }
}

LRT::~LRT(){
  if(pDat != 0)
    delete[] pDat;
  if(pLRT != 0)
    delete[] pLRT;

  if(pQ != 0){
    for(int i = 0; i < n - 2*order - 1; i++)
      if(pQ[i] != 0)
        delete[] pQ[i];
  delete[] pQ;
  }
}

void LRT::testStatForward(){
  //initialize XTXInv, XTx and xTx for computing Q(p + 1, tau)
  Matrix XTXInvLow = initX(2*order);
  Vector XTxLow = initXTx(XTXInvLow, order, 2*order);
  //now compute XTX
  XTXInvLow = XTXInvLow.trans()*XTXInvLow;
  //invert XTX
  XTXInvLow = XTXInvLow.cholInv();
  double xTxLow = initxTx(order, 2*order);

  //initialize updating vectors with all unit elements
  Vector xLow(order + 1, 1.);

  for(int t = (2*order + 1); t < n; t++){
    //update XTXInv, XTx and xTx for Q(p + 1, t)
    updateXVec(xLow, t);
    updateXTXinv(XTXInvLow, xLow);
    updateXTx(XTxLow, t);
    xTxLow += pDat[t]*pDat[t];
    pQ[t - 2*order - 1][0] = xTxLow - XTxLow.dotProd(XTXInvLow*XTxLow);
  }

}

void LRT::testStatBackward(){
  //initialize XTXInv, XTx and xTx for computing Q(tau + 1, n)
  Matrix XTXInvUp = initX(n - 1);
  Vector XTxUp = initXTx(XTXInvUp, n - order - 1, n- 1);
  XTXInvUp = XTXInvUp.trans()*XTXInvUp;
  XTXInvUp = XTXInvUp.cholInv();
  double xTxUp = initxTx(n - order - 1, n - 1);

  //initialize updating vectors with all unit elements
  Vector xUp(order + 1, 1.);

  for(int t = (2*order + 1); t < n; t++){
    //update XTXInv, XTx and xTx for Q(t + 1, n)
    updateXVec(xUp, n - t - 1 + order);
    updateXTXinv(XTXInvUp, xUp);
    updateXTx(XTxUp, n - t - 1 + order);
    xTxUp += pDat[n - t - 1 + order]*pDat[n - t - 1 + order];

    pQ[t - 2*order - 1][1] = xTxUp - XTxUp.dotProd(XTXInvUp*XTxUp);
  }
}

void LRT::testStat(){
  pLRT = new double[n - 3*order - 3];
  memoryCheck(pLRT);

  //combine the element of Q to obtain the raw LRT statistics
  for(int i = 2*(order + 1); i <= (n - order - 2); i++)
    pLRT[i - 2*(order + 1)] = pQ[n - 2*(order + 1)][0] 
      - pQ[i - 2*(order + 1)][0] - pQ[n - i - order - 2][1];
}

Matrix LRT::initX(int t){
  double** pA = new double*[order + 1];

  for(int i=0; i <= order; i++){
    pA[i] = new double[order + 1];
    pA[i][0] = 1.;
    for(int j = 1; j <= order; j++){
      pA[i][j] = pDat[t - order + i - j];
    }
  }
  Matrix A(order + 1, order + 1, pA);
  
  for(int i=0; i <= order; i++)
    delete[] pA[i];
  delete[] pA;

  return A;
}

Vector LRT::initXTx(Matrix& A, int low, int up){
  Vector temp(pDat, low, up);
  return A.trans()*temp;
}

double LRT::initxTx(int low, int up){
  //begin with all unit elements
  double a = 0;
  for(int i=low; i <= up; i++)
    a += pDat[i]*pDat[i];
  return a;
}

void LRT::updateXTXinv(Matrix& A, Vector& x){
  Vector temp = A*x;
  double d = 1. + x.dotProd(temp);
  A -= (temp*temp)*(1./d);
}

void LRT::updateXTx(Vector& a, int t){
  //assumes a[0] = 1.
  a.pA[0] += pDat[t];
  for(int i = 1; i <= order; i++)
    a.pA[i] += pDat[t - i]*pDat[t];
}

void LRT::updateXVec(Vector& x, int t){
  //assumes pA[0] = 1.
  for(int i = 1; i <= order; i++)
    x.pA[i] = pDat[t - i];
}

void LRT::memoryCheck(double** p){
  if(p == 0){
    std::cout << "Memory allocation for pointer-to-pointer failed" << std::endl;
    exit(1);
  }
}

void LRT::memoryCheck(double* p){
  if(p == 0){
    std::cout << "Memory allocation failed" << std::endl;
    exit(1);
  }
}

